-module(schedule).

-behaviour(gen_server).

-export([start_link/0, init/1, handle_cast/2, handle_call/3]).

-export([set_test_config/0]).

-include("erlk.hrl").

-record(division, {teams = [], practice = 0, qualification = 0}).
-record(config, {divisions = []}).
-record(state, {database_file = none, database = none, teams = [], config = none}).

start_link() ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

init([]) ->
  rand:seed(default),
  {ok, #state{}}.

duplicate(1, Original, Constructed) -> lists:append(Original, Constructed);
duplicate(N, Original, Constructed) -> duplicate(N-1, Original, lists:append(Original, Constructed)).

duplicate(List, N) -> duplicate(List, N, []).

padding(Size) when Size >= 4 -> (4 - (Size rem 4)) rem 4.

fill_padding_match([], Teams, Filled) -> {Filled, Teams};
fill_padding_match([padding | Left], [Picked | Teams], Filled) ->
  Duplicate = lists:member(Picked, Left) or lists:member(Picked, Filled),
  if Duplicate -> fill_padding_match([padding | Left], lists:append(Teams, [Picked]), Filled);
     true -> fill_padding_match(Left, Teams, [Picked | Filled])
  end;
fill_padding_match([Team | Left], Teams, Filled) ->
  fill_padding_match(Left, Teams, [Team | Filled]).

fill_padding_match(Match, Teams) ->
  fill_padding_match(Match, Teams, []).

fill_padding_schedule(Matches, Teams) ->
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- Teams])],
  {Schedule, _} = lists:mapfoldl(fun fill_padding_match/2, TeamsRandom, Matches),
  Schedule.

init_db(Database) ->
  ok = sqlite3:create_table(Database, teams, [{name, text, [not_null]},
                                              {inspection, blob}],
                            [{primary_key, [name]}]),
  ok = sqlite3:create_table(Database, matches, [{division, integer, [not_null]},
                                                {round, text, [not_null]},
                                                {number, integer, [not_null]},
                                                {genid, text, [not_null]},
                                                {red_1, text, [not_null]},
                                                {red_2, text, [not_null]},
                                                {blue_1, text, [not_null]},
                                                {blue_2, text, [not_null]}],
                            [{primary_key, [division, round, number, genid]},
                             {check, "round IN ('practice', 'qualification', 'elimination')"},
                             {foreign_key, {[blue_1], teams, [name], "ON DELETE RESTRICT"}},
                             {foreign_key, {[blue_2], teams, [name], "ON DELETE RESTRICT"}},
                             {foreign_key, {[red_1], teams, [name], "ON DELETE RESTRICT"}},
                             {foreign_key, {[red_2], teams, [name], "ON DELETE RESTRICT"}}]),

  ok = sqlite3:create_table(Database, match_scores, [{division, integer, [not_null]},
                                                     {round, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {genid, text, [not_null]},
                                                     {instance, integer, [not_null]},
                                                     {score, blob, [not_null]}],
                            [{primary_key, [division, round, number, genid, instance]},
                             {foreign_key, {[division, round, number, genid], matches, [division, round, number, genid], "ON DELETE CASCADE"}}]),
  ok = sqlite3:create_table(Database, match_states, [{division, integer, [not_null]},
                                                     {round, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {genid, text, [not_null]},
                                                     {time, integer, [not_null]},
                                                     {state, blob, [not_null]}],
                            [{primary_key, [division, round, number, genid, time]},
                             {foreign_key, {[division, round, number, genid], matches, [division, round, number, genid], "ON DELETE CASCADE"}}]),

  % TODO: finals
  % ok = sqlite3:create_table(Database, finals, [], [{primary_key, []}]),
  ok = sqlite3:create_table(Database, skills_scores, [{team, text, [not_null]},
                                                      {type, text, [not_null]},
                                                      {attempt, integer, [not_null]},
                                                      {score, blob, [not_null]}],
                            [{primary_key, [team, type, attempt]},
                             {foreign_key, {[team], teams, [name], "ON DELETE CASCADE"}}]),
  ok.

delete_teams(Database, Teams) ->
  SQL = lists:append(["BEGIN; ",
         lists:append([io_lib:format("DELETE FROM teams WHERE name = ~p; ",
                                     [Team]) || Team <- Teams]),
         "COMMIT;"]),
  sqlite3:sql_exec_script(Database, SQL).

add_teams(Database, Teams) ->
  SQL = lists:append(["BEGIN; ",
         lists:append([io_lib:format("INSERT INTO teams(name) VALUES('~s') ON CONFLICT(name) DO NOTHING; ",
                                     [Team]) || Team <- Teams]),
         "COMMIT;"]),
  sqlite3:sql_exec_script(Database, SQL).

first_error([]) -> ok;
first_error([ok | Rest]) -> first_error(Rest);
first_error([Error | _]) -> Error.

delete_matches(Database, Division, Round, GenID) ->
  SQL = io_lib:format("DELETE FROM matches WHERE division = ~p AND genid = ~p AND round = '~p'", [Division, GenID, Round]),
  sqlite3:sql_exec_script(Database, SQL).

write_matches(Database, Division, Round, GenID, Matches) ->
  Schedule = [[{division, Division},
               {round, atom_to_list(Round)},
               {number, N},
               {genid, GenID},
               {blue_1, B1},
               {blue_2, B2},
               {red_1, R1},
               {red_2, R2}] || {N, [B1, B2, R1, R2]} <- lists:enumerate(Matches)],
  [ok | Resp] = sqlite3:write_many(Database, matches, Schedule),
  lists:last(Resp).

create_schedule(Seed, Matches, Teams) ->
  TeamsRepeated = duplicate(Matches, Teams),
  TeamsPadded = TeamsRepeated ++ lists:duplicate(padding(length(Teams)*Matches), padding),
  rand:seed(default, Seed),
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- TeamsPadded])],
  fill_padding_schedule(pick_matches(TeamsRandom), Teams).

pick_teams(Teams, Picked, 0, Safety) when Safety > 0 -> {Teams, Picked};
pick_teams([Team | Teams], Picked, N, Safety) when Safety > 0 ->
  Duplicate = lists:member(Team, Picked),
  if Duplicate, Team =/= padding -> pick_teams(lists:append(Teams, [Team]), Picked, N, Safety - 1);
     true -> pick_teams(Teams, [Team | Picked], N-1, Safety)
  end.

pick_teams(Teams, N) -> pick_teams(Teams, [], N, 1000).

pick_matches([], Matches) -> Matches;
pick_matches(Teams, Matches) ->
  {NewTeams, Match} = pick_teams(Teams, 4),
  pick_matches(NewTeams, [Match | Matches]).

pick_matches(Teams) -> pick_matches(Teams, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-define(TEST_CONFIG, #config{divisions = [#division{teams = ["A", "B", "C", "D"], practice = 1}]}).
set_test_config() ->
  gen_server:call(?MODULE, {set_config, ?TEST_CONFIG}).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_call({set_config, Config}, _, State) when is_record(Config, config) ->
  {reply, ok, State#state{config = Config}};
handle_call({new_schedule, Division, Round, Seed}, _, State) ->
  DivInfo = lists:nth(Division, State#state.config#config.divisions),
  Teams = DivInfo#division.teams,
  Matches = case Round of
      practice when DivInfo#division.practice =/= 0 ->
        DivInfo#division.practice;
      qualification when DivInfo#division.qualification =/= 0->
        DivInfo#division.qualification
  end,
  MatchTeams = create_schedule(Seed, Matches, Teams),
  GenID = lists:sublist([ Y || <<X:4>> <= crypto:hash(sha, Teams ++ [Seed, Matches]),
                               Y <- integer_to_list(X, 16)], 4),
  ok = write_matches(State#state.database, Division, Round, GenID, MatchTeams),
  {reply, GenID, State};
handle_call({new_db, DatabaseFile}, _, State) ->
  ok = if State#state.database =:= none -> ok;
          true -> sqlite3:close(State#state.database)
       end,
  Exists = filelib:is_regular(DatabaseFile),
  ok = if Exists -> file:delete(DatabaseFile);
          true -> ok
       end,
  {ok, Database} = sqlite3:open(schedule_db, [{file, DatabaseFile}]),
  ok = sqlite3:sql_exec(Database, "PRAGMA foreign_keys = ON;"),
  {reply, init_db(Database), State#state{database_file = DatabaseFile, database = Database}};
handle_call({load_db, DatabaseFile}, _, State) ->
  true = filelib:is_regular(DatabaseFile),
  {ok, Database} = if DatabaseFile == State#state.database_file ->
                        {ok, State#state.database};
                      true ->
                        ok = if State#state.database =:= none -> ok;
                                true -> sqlite3:close(State#state.database)
                             end,
                        {ok, DB} = sqlite3:open(schedule_db, [{file, DatabaseFile}]),
                        {sqlite3:sql_exec(DB, "PRAGMA foreign_keys = ON;"), DB}
                   end,
  {reply, ok, State#state{database_file = DatabaseFile, database = Database}}.

handle_cast({delete_teams, Removed}, State) ->
  Teams = lists:filter(fun(X) -> lists:member(X, Removed) =:= false end, State#state.teams),
  ok = first_error(delete_teams(State#state.database, Removed)),
  {noreply, State#state{teams = Teams}};
handle_cast({add_teams, Teams}, State) ->
  ok = first_error(add_teams(State#state.database, Teams)),
  {noreply, State#state{teams = Teams}};
handle_cast({delete_schedule, Division, Round, GenID}, State) ->
  ok = first_error(delete_matches(State#state.database, Division, Round, GenID)),
  {noreply, State}.
